"""Data types and converters."""

# The MIT License (MIT)
#
# Copyright (c) 2015 by Teradata
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import re
import json
from . import util
from .api import *  # @UnusedWildImport # noqa

logger = logging.getLogger(__name__)

SECS_IN_MILLISECS = MILLISECS_IN_MICROSECS = 1000
dateRegExStr = r"(\d{4})-(\d{2})-(\d{2})"
timeRegExStr = r"(\d{2}):(\d{2}):(\d{2})(\.(\d{1,6}))?(([-+])(\d{2}):(\d{2}))?"
dateRegEx = re.compile("^{}$".format(dateRegExStr))
timeRegEx = re.compile("^{}$".format(timeRegExStr))
timestampRegEx = re.compile("^{} {}$".format(dateRegExStr, timeRegExStr))
scalarIntervalRegEx = re.compile("^(-?)(\d+)$")
yearToMonthIntervalRegEx = re.compile("^(-?)(\d+)-(\d+)$")
dayToHourIntervalRegEx = re.compile("^(-?)(\d+) (\d+)$")
dayToMinuteIntervalRegEx = re.compile("^(-?)(\d+) (\d+):(\d+)$")
dayToSecondIntervalRegEx = re.compile("^(-?)(\d+) (\d+):(\d+):(\d+\.?\d*)$")
hourToMinuteIntervalRegEx = re.compile("^(-?)(\d+):(\d+)$")
hourToSecondIntervalRegEx = re.compile("^(-?)(\d+):(\d+):(\d+\.?\d*)$")
minuteToSecondIntervalRegEx = re.compile("^(-?)(\d+):(\d+\.?\d*)$")
secondIntervalRegEx = re.compile("^(-?)(\d+\.?\d*)$")
periodRegEx1 = re.compile("\('(.*)',\s*'(.*)'\)")
periodRegEx2 = re.compile("ResultStruct:PERIOD\(.*\)\[(.*),\s*(.*)\]")

NUMBER_TYPES = ("BYTEINT", "BIGINT", "DECIMAL", "DOUBLE", "DOUBLE PRECISION",
                "INTEGER", "NUMBER", "SMALLINT", "FLOAT", "INT", "NUMERIC",
                "REAL")

INT_TYPES = ("BYTEINT", "BIGINT", "INTEGER", "SMALLINT", "INT")

FLOAT_TYPES = ("FLOAT", "DOUBLE", "DOUBLE PRECISION", "REAL")

BINARY_TYPES = ("BLOB", "BYTE", "VARBYTE")


def _getMs(m, num):
    ms = m.group(num)
    if ms:
        ms = int(ms.ljust(6, "0"))
    else:
        ms = 0
    return ms


def _getInt(m, num):
    return int(m.group(num))


def _getFloat(m, num):
    return float(m.group(num))


def convertDate(value):
    m = dateRegEx.match(value)
    if m:
        return datetime.date(_getInt(m, 1), _getInt(m, 2), _getInt(m, 3))
    else:
        raise InterfaceError(
            "INVALID_DATE", "Date format invalid: {}".format(value))


def convertTime(value):
    m = timeRegEx.match(value)
    if m:
        tz = None
        if m.group(7):
            tz = TimeZone(m.group(7), _getInt(m, 8), _getInt(m, 9))
        return datetime.time(_getInt(m, 1), _getInt(m, 2), _getInt(m, 3),
                             _getMs(m, 5), tz)
    else:
        raise InterfaceError(
            "INVALID_TIME", "Time format invalid: {}".format(value))


def convertTimestamp(value):
    m = timestampRegEx.match(value)
    if m:
        tz = None
        if m.group(10):
            tz = TimeZone(m.group(10), _getInt(m, 11), _getInt(m, 12))
        return datetime.datetime(_getInt(m, 1), _getInt(m, 2), _getInt(m, 3),
                                 _getInt(m, 4), _getInt(m, 5), _getInt(m, 6),
                                 _getMs(m, 8), tz)
    else:
        raise InterfaceError(
            "INVALID_TIMESTAMP", "Timestamp format invalid: {}".format(value))


def _convertScalarInterval(dataType, value, *args):
    return _convertInterval(dataType, value, scalarIntervalRegEx, *args)


def _convertInterval(dataType, value, regEx, *args):
    m = regEx.match(value)
    if m:
        kwargs = {}
        index = 2
        for field in args:
            if field != "seconds":
                kwargs[field] = _getInt(m, index)
            else:
                kwargs[field] = _getFloat(m, index)
            index += 1
        return Interval(negative=True if m.group(1) else False, **kwargs)
    else:
        raise InterfaceError("INVALID_INTERVAL",
                             "{} format invalid: {}".format(dataType, value))


def convertInterval(dataType, value):
    value = value.strip()
    if dataType == "INTERVAL YEAR":
        return _convertScalarInterval(dataType, value, "years")
    elif dataType == "INTERVAL YEAR TO MONTH":
        return _convertInterval(dataType, value, yearToMonthIntervalRegEx,
                                "years", "months")
    elif dataType == "INTERVAL MONTH":
        return _convertScalarInterval(dataType, value, "months")
    elif dataType == "INTERVAL DAY":
        return _convertScalarInterval(dataType, value, "days")
    elif dataType == "INTERVAL DAY TO HOUR":
        return _convertInterval(dataType, value, dayToHourIntervalRegEx,
                                "days", "hours")
    elif dataType == "INTERVAL DAY TO MINUTE":
        return _convertInterval(dataType, value, dayToMinuteIntervalRegEx,
                                "days", "hours", "minutes")
    elif dataType == "INTERVAL DAY TO SECOND":
        return _convertInterval(dataType, value, dayToSecondIntervalRegEx,
                                "days", "hours", "minutes", "seconds")
    elif dataType == "INTERVAL HOUR":
        return _convertScalarInterval(dataType, value, "hours")
    elif dataType == "INTERVAL HOUR TO MINUTE":
        return _convertInterval(dataType, value, hourToMinuteIntervalRegEx,
                                "hours", "minutes")
    elif dataType == "INTERVAL HOUR TO SECOND":
        return _convertInterval(dataType, value, hourToSecondIntervalRegEx,
                                "hours", "minutes", "seconds")
    elif dataType == "INTERVAL MINUTE":
        return _convertScalarInterval(dataType, value, "minutes")
    elif dataType == "INTERVAL MINUTE TO SECOND":
        return _convertInterval(dataType, value, minuteToSecondIntervalRegEx,
                                "minutes", "seconds")
    elif dataType == "INTERVAL SECOND":
        return _convertInterval(dataType, value, secondIntervalRegEx,
                                "seconds")
    return value


def convertPeriod(dataType, value):
    m = periodRegEx1.match(value)
    if not m:
        m = periodRegEx2.match(value)
    if m:
        if "TIMESTAMP" in dataType:
            start = convertTimestamp(m.group(1))
            end = convertTimestamp(m.group(2))
        elif "TIME" in dataType:
            start = convertTime(m.group(1))
            end = convertTime(m.group(2))
        elif "DATE" in dataType:
            start = convertDate(m.group(1))
            end = convertDate(m.group(2))
        else:
            raise InterfaceError("INVALID_PERIOD",
                                 "Unknown PERIOD data type: {}".format(
                                     dataType, value))
    else:
        raise InterfaceError(
            "INVALID_PERIOD", "{} format invalid: {}".format(dataType, value))
    return Period(start, end)


def zeroIfNone(value):
    if value is None:
        value = 0
    return value


class DataTypeConverter:

    """Handles conversion of result set data types into python objects."""

    def convertValue(self, dbType, dataType, typeCode, value):
        """Converts the value returned by the database into the desired
         python object."""
        raise NotImplementedError(
            "convertValue must be implemented by sub-class")

    def convertType(self, dbType, dataType):
        """Converts the data type to a python type code."""
        raise NotImplementedError(
            "convertType must be implemented by sub-class")


class DefaultDataTypeConverter (DataTypeConverter):

    """Handles conversion of result set data types into python objects."""

    def __init__(self, useFloat=False):
        self.useFloat = useFloat

    def convertValue(self, dbType, dataType, typeCode, value):
        """Converts the value returned by the database into the desired
         python object."""
        logger.trace(
            "Converting \"%s\" to (%s, %s).", value, dataType, typeCode)
        if value is not None:
            if typeCode == NUMBER:
                try:
                    return NUMBER(value)
                except:
                    # Handle infinity and NaN for older ODBC drivers.
                    if value == "1.#INF":
                        return NUMBER('Infinity')
                    elif value == "-1.#INF":
                        return NUMBER('-Infinity')
                    else:
                        return NUMBER('NaN')
            elif typeCode == float:
                return value if not util.isString else float(value)
            elif typeCode == Timestamp:
                if util.isString(value):
                    return convertTimestamp(value)
                else:
                    return datetime.datetime.fromtimestamp(
                        value // SECS_IN_MILLISECS).replace(
                        microsecond=value % SECS_IN_MILLISECS *
                        MILLISECS_IN_MICROSECS)
            elif typeCode == Time:
                if util.isString(value):
                    return convertTime(value)
                else:
                    return datetime.datetime.fromtimestamp(
                        value // SECS_IN_MILLISECS).replace(
                        microsecond=value % SECS_IN_MILLISECS *
                        MILLISECS_IN_MICROSECS).time()
            elif typeCode == Date:
                if util.isString(value):
                    return convertDate(value)
                else:
                    return datetime.datetime.fromtimestamp(
                        value // SECS_IN_MILLISECS).replace(
                        microsecond=value % SECS_IN_MILLISECS *
                        MILLISECS_IN_MICROSECS).date()
            elif typeCode == BINARY:
                if util.isString(value):
                    return bytearray.fromhex(value)
            elif dataType.startswith("INTERVAL"):
                return convertInterval(dataType, value)
            elif dataType.startswith("JSON") and util.isString(value):
                return json.loads(value, parse_int=decimal.Decimal,
                                  parse_float=decimal.Decimal)
            elif dataType.startswith("PERIOD"):
                return convertPeriod(dataType, value)
        return value

    def convertType(self, dbType, dataType):
        """Converts the data type to a python type code."""
        typeCode = STRING
        if dataType in NUMBER_TYPES:
            typeCode = NUMBER
            if self.useFloat and dataType in FLOAT_TYPES:
                typeCode = float
        elif dataType in BINARY_TYPES:
            typeCode = BINARY
        elif dataType.startswith("DATE"):
            typeCode = Date
        elif dataType.startswith("TIMESTAMP"):
            typeCode = Timestamp
        elif dataType.startswith("TIME"):
            typeCode = Time
        return typeCode


class TimeZone (datetime.tzinfo):

    """Represents a Fixed Time Zone offset from UTC."""

    def __init__(self, sign, hours, minutes):
        self.offset = datetime.timedelta(hours=hours, minutes=minutes)
        if sign == "-":
            self.offset = -self.offset

    def utcoffset(self, dt):
        return self.offset

    def tzname(self, dt):
        return "TeradataTimestamp"

    def dst(self, dt):
        return 0


def _appendInterval(arr, value, padding=2, separator=" "):
    if value is not None:
        if arr and separator:
            arr.append(separator)
        s = (("%0" + str(padding + 7) + ".6f") % value).rstrip("0").rstrip(".")
        arr.append(s)


class Interval:

    """Represents a SQL date/time interval."""

    def __init__(self, negative=False, years=None, months=None, days=None,
                 hours=None, minutes=None, seconds=None):
        self.negative = negative
        self.years = years
        self.months = months
        self.days = days
        self.hours = hours
        self.minutes = minutes
        self.seconds = seconds
        self.type = None
        if years is not None:
            if months is not None:
                self.type = "YEAR TO MONTH"
            else:
                self.type = "YEAR"
            if days or hours or minutes or seconds:
                raise InterfaceError(
                    "INVALID INTERVAL",
                    "A year/month interval cannot be "
                    "shared with a day/hour/minute/second interval.")
        elif months is not None:
            self.type = "MONTH"
            if days or hours or minutes or seconds:
                raise InterfaceError(
                    "INVALID INTERVAL",
                    "A year/month interval cannot be shared "
                    "with a day/hour/minute/second interval.")
        elif days is not None:
            if seconds is not None:
                self.type = "DAY TO SECOND"
                self.hours = zeroIfNone(hours)
                self.minutes = zeroIfNone(minutes)
            elif minutes is not None:
                self.type = "DAY TO MINUTE"
                self.hours = zeroIfNone(hours)
            elif hours is not None:
                self.type = "DAY TO HOUR"
            else:
                self.type = "DAY"
        elif hours is not None:
            if seconds is not None:
                self.type = "HOUR TO SECOND"
                self.minutes = zeroIfNone(minutes)
            elif minutes is not None:
                self.type = "HOUR TO MINUTE"
            elif hours is not None:
                self.type = "DAY TO HOUR"
            else:
                self.type = "HOUR"
        elif minutes is not None:
            if seconds is not None:
                self.type = "MINUTE TO SECOND"
            else:
                self.type = "MINUTE"
        elif seconds is not None:
            self.type = "SECOND"
        else:
            raise InterfaceError(
                "INVALID INTERVAL",
                "One of years, months, days, hours, minutes, "
                "seconds must not be None.")

    def timedelta(self):
        if self.years or self.months:
            raise InterfaceError(
                "UNSUPPORTED_INTERVAL",
                "timedelta() is not supported for Year "
                "and Month interval types. %s" % repr(self))
        delta = datetime.timedelta(days=zeroIfNone(self.days),
                                   hours=zeroIfNone(self.hours),
                                   minutes=zeroIfNone(self.minutes),
                                   seconds=zeroIfNone(self.seconds))
        if self.negative:
            delta = -delta
        return delta

    def __str__(self):
        s = []
        _appendInterval(s, self.years, padding=1)
        _appendInterval(s, self.months, separator="-")
        _appendInterval(s, self.days, padding=1)
        _appendInterval(s, self.hours)
        _appendInterval(s, self.minutes, separator=":")
        _appendInterval(s, self.seconds, separator=":")
        if self.negative:
            s.insert(0, "-")
        return "".join(s)

    def __repr__(self):
        return str(self.__dict__)

    def __eq__(self, other):
        try:
            return self.__dict__ == other.__dict__
        except AttributeError:
            return False

    def __ne__(self, other):
        return not self == other


class Period:

    """ Represents a PERIOD data type. """

    def __init__(self, start, end):
        self.start = start
        self.end = end

    def __str__(self):
        return "('" + str(self.start) + "', '" + str(self.end) + "')"

    def __eq__(self, other):
        try:
            return self.__dict__ == other.__dict__
        except AttributeError:
            return False

    def __ne__(self, other):
        return not self == other



import teradata

udaExec = teradata.UdaExec(appName="HelloWorld", version="1.0",
                           logConsole=False)

#session = udaExec.connect(method="odbc", system="tdprod",
#        username="xxx", password="xxx");

#for row in session.execute("SELECT GetQueryBand()"):
#    print(row)



"""An implementation of the Python Database API Specification v2.0 using
Teradata REST."""

# The MIT License (MIT)
#
# Copyright (c) 2015 by Teradata
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import atexit
import base64
import json
import ssl
import sys
import time
import io

from . import pulljson, util, datatypes
from .api import *  # @UnusedWildImport # noqa


if sys.version_info[0] == 2:
    import httplib as httplib  # @UnresolvedImport #@UnusedImport
else:
    import http.client as httplib  # @UnresolvedImport @UnusedImport @Reimport
    unicode = str

logger = logging.getLogger(__name__)

REST_ERROR = "REST_ERROR"
HTTP_STATUS_DATABASE_ERROR = 420
ERROR_USER_GENERATED_TRANSACTION_ABORT = 3514
MAX_CONNECT_RETRIES = 5

connections = []


def cleanup():
    for conn in connections:
        conn.close()
atexit.register(cleanup)


class RestConnection:

    """ Represents a Connection to Teradata using the REST API for
     Teradata Database """

    def __init__(self, dbType="Teradata", host=None, system=None,
                 username=None, password=None, protocol='http', port=None,
                 webContext='/tdrest', autoCommit=False, implicit=False,
                 transactionMode='TERA', queryBands=None, charset=None,
                 verifyCerts=True, sslContext=None, database=None,
                 authentication=None,
                 dataTypeConverter=datatypes.DefaultDataTypeConverter()):
        self.dbType = dbType
        self.system = system
        self.sessionId = None
        self.implicit = implicit
        self.transactionMode = transactionMode
        self.dataTypeConverter = dataTypeConverter
        self.cursors = []
        # Support TERA and Teradata as transaction mode to be consistent with
        # ODBC.
        if transactionMode == "Teradata":
            self.transactionMode = "TERA"
        self.autoCommit = False
        if port is None:
            if protocol == 'http':
                port = 1080
            elif protocol == 'https':
                port = 1443
            else:
                raise InterfaceError(
                    CONFIG_ERROR, "Unsupported protocol: {}".format(protocol))
        if host is None:
            raise InterfaceError(0,
                                 "\"host\" is a required field, "
                                 "set to location of "
                                 "TDREST server.")
        self.template = RestTemplate(
            protocol, host, int(port), webContext, username, password,
            accept='application/vnd.com.teradata.rest-v1.0+json',
            verifyCerts=util.booleanValue(verifyCerts), sslContext=sslContext)
        with self.template.connect() as conn:
            if not self.implicit:
                options = {}
                options['autoCommit'] = autoCommit
                options['transactionMode'] = transactionMode
                if queryBands:
                    options['queryBands'] = queryBands
                if charset:
                    options['charSet'] = charset
                if database:
                    options['defaultDatabase'] = database
                if authentication:
                    options['logMech'] = authentication
                try:
                    session = conn.post(
                        '/systems/{0}/sessions'.format(self.system),
                        options).readObject()
                    self.sessionId = session['sessionId']
                    connections.append(self)
                    logger.info("Created explicit session: %s",  session)
                except (pulljson.JSONParseError) as e:
                    raise InterfaceError(
                        e.code, "Error reading JSON response: " + e.msg)

    def close(self):
        """ Closes an Explicit Session using the REST API for Teradata
         Database """
        if hasattr(self, 'sessionId') and self.sessionId is not None:
            with self.template.connect() as conn:
                try:
                    conn.delete(
                        '/systems/{0}/sessions/{1}'.format(
                            self.system, self.sessionId))
                except InterfaceError as e:
                    # Ignore if the session is already closed.
                    if e.code != 404:
                        raise
            logger.info("Closing session: %s", self.sessionId)
            self.sessionId = None
            connections.remove(self)
        for cursor in list(self.cursors):
            cursor.close()

    def commit(self):
        with self.cursor() as cursor:
            if self.transactionMode == 'ANSI':
                cursor.execute("COMMIT")
            else:
                cursor.execute("ET")

    def rollback(self):
        with self.cursor() as cursor:
            try:
                cursor.execute("ROLLBACK")
            except DatabaseError as e:
                if e.code != ERROR_USER_GENERATED_TRANSACTION_ABORT:
                    raise

    def cursor(self):
        return RestCursor(self)

    def __del__(self):
        self.close()

    def __enter__(self):
        return self

    def __exit__(self, t, value, traceback):
        self.close()

connect = RestConnection


class RestCursor (util.Cursor):

    def __init__(self, connection):
        self.conn = None
        util.Cursor.__init__(
            self, connection, connection.dbType, connection.dataTypeConverter)
        self.conn = connection.template.connect()
        connection.cursors.append(self)

    def callproc(self, procname, params, queryTimeout=None):
        inparams = None
        outparams = None
        count = 0
        query = "CALL {} (".format(procname)
        if params is not None:
            inparams = [[]]
            outparams = []
            for p in params:
                if count > 0:
                    query += ", "
                if isinstance(p, InOutParam):
                    inparams[0].append(p.inValue)
                    outparams.append(p.inValue)
                elif isinstance(p, OutParam):
                    outparams.append(None)
                else:
                    inparams[0].append(p)
                count += 1
                query += "?"
        query += ")"
        outparams = self._handleResults(self._execute(
            query, inparams, outparams, queryTimeout=queryTimeout),
            len(outparams) > 0)
        return util.OutParams(params,  self.dbType, self.converter, outparams)

    def close(self):
        if self.conn:
            self.conn.close()

    def execute(self, query, params=None, queryTimeout=None):
        if params is not None:
            params = [params]
        self._handleResults(
            self._execute(query, params, queryTimeout=queryTimeout))
        return self

    def executemany(self, query, params, batch=False, queryTimeout=None):
        self._handleResults(
            self._execute(query, params, batch=batch,
                          queryTimeout=queryTimeout))
        return self

    def _handleResults(self, results, hasOutParams=False):
        self.results = results
        try:
            results.expectObject()
            self.queueDuration = results.expectField(
                "queueDuration", pulljson.NUMBER)
            self.queryDuration = results.expectField(
                "queryDuration", pulljson.NUMBER)
            logger.debug("Durations reported by REST service: Queue Duration: "
                         "%s, Query Duration: %s", self.queueDuration,
                         self.queryDuration)
            results.expectField("results", pulljson.ARRAY)
            results.expectObject()
            return self._handleResultSet(results, hasOutParams)
        except (pulljson.JSONParseError) as e:
            raise InterfaceError(
                e.code, "Error reading JSON response: " + e.msg)

    def _execute(self, query, params=None, outParams=None, batch=False,
                 queryTimeout=None):
        options = {}
        options['query'] = query
        options['format'] = 'array'
        options['includeColumns'] = 'true'
        options['rowLimit'] = 0
        if params is not None:
            options['params'] = list(
                list(_convertParam(p) for p in paramSet)
                for paramSet in params)
            options['batch'] = batch
        if outParams is not None:
            options['outParams'] = outParams
        if not self.connection.implicit:
            options['session'] = str(self.connection.sessionId)
        if queryTimeout is not None:
            options['queryTimeout'] = queryTimeout
            options['queueTimeout'] = queryTimeout
        return self.conn.post('/systems/{0}/queries'.format(
            self.connection.system), options)

    def _handleResultSet(self, results, hasOutParams=False):
        outParams = None
        if hasOutParams:
            outParams = results.expectField(
                "outParams", pulljson.ARRAY, readAll=True)
            self.resultSet = None
        else:
            try:
                self.resultSet = results.expectField(
                    "resultSet", pulljson.BOOLEAN)
            except pulljson.JSONParseError:
                # Workaround for Batch mode and Stored procedures which doens't
                # include a resultSet.
                self.resultSet = None
        if self.resultSet:
            index = 0
            self.columns = {}
            self.description = []
            self.types = []
            self.rowcount = -1
            self.rownumber = None
            for column in results.expectField("columns", pulljson.ARRAY):
                self.columns[column["name"].lower()] = index
                type_code = self.converter.convertType(
                    self.dbType, column["type"])
                self.types.append((column["type"], type_code))
                self.description.append(
                    (column["name"], type_code, None, None, None, None, None))
                index += 1
            self.iterator = results.expectField("data", pulljson.ARRAY)
        else:
            self.columns = None
            self.description = None
            self.rownumber = None
            self.rowcount = -1
            if self.resultSet is not None:
                self.rowcount = results.expectField("count")
        return outParams

    def nextset(self):
        for row in self:  # @UnusedVariable
            pass
        for event in self.results:
            if event.type == pulljson.START_OBJECT:
                self._handleResultSet(self.results)
                return True


def _convertParam(p):
    if util.isString(p) or p is None:
        return p
    elif isinstance(p, bytearray):
        return ''.join('{:02x}'.format(x) for x in p)
    else:
        return unicode(p)


class RestTemplate:

    def __init__(self, protocol, host, port, webContext, username, password,
                 sslContext=None, verifyCerts=True, accept=None):
        self.protocol = protocol
        self.host = host
        self.port = port
        self.webContext = webContext
        self.headers = {}
        self.headers['Content-Type'] = 'application/json'
        if accept is not None:
            self.headers['Accept'] = accept
        self.headers['Authorization'] = 'Basic ' + \
            base64.b64encode(
                (username + ":" + password).encode('utf_8')).decode('ascii')
        self.sslContext = sslContext
        if sslContext is None and not verifyCerts:
            self.sslContext = ssl.create_default_context()
            self.sslContext.check_hostname = False
            self.sslContext.verify_mode = ssl.CERT_NONE

    def connect(self):
        return HttpConnection(self)


class HttpConnection:

    def __init__(self, template):
        self.template = template
        if template.protocol.lower() == "http":
            self.conn = httplib.HTTPConnection(template.host, template.port)
        elif template.protocol.lower() == "https":
            self.conn = httplib.HTTPSConnection(
                template.host, template.port, context=template.sslContext)
        else:
            raise InterfaceError(
                REST_ERROR, "Unknown protocol: %s" % template.protocol)
        failureCount = 0
        while True:
            try:
                self.conn.connect()
                break
            except Exception as e:
                eofError = "EOF occurred in violation of protocol" in str(e)
                failureCount += 1
                if not eofError or failureCount > MAX_CONNECT_RETRIES:
                    raise InterfaceError(
                        REST_ERROR,
                        "Error accessing {}:{}. ERROR:  {}".format(
                            template.host, template.port, e))
                else:
                    logger.debug(
                        "Received an \"EOF occurred in violation of "
                        "protocol\" error, retrying connection.")

    def close(self):
        if self.conn:
            self.conn.close()

    def post(self, uri, data={}):
        return self.send(uri, 'POST', data)

    def delete(self, uri):
        self.send(uri, 'DELETE', None)

    def get(self, uri):
        return self.send(uri, 'GET', None)

    def __enter__(self):
        return self

    def __exit__(self, t, value, traceback):
        self.close()

    def send(self, uri, method, data):
        response = None
        url = self.template.webContext + uri
        try:
            start = time.time()
            payload = json.dumps(data).encode('utf8') if data else None
            logger.trace("%s: %s, %s", method, url, payload)
            self.conn.request(method, url, payload, self.template.headers)
            response = self.conn.getresponse()
            duration = time.time() - start
            logger.debug("Roundtrip Duration: %.3f seconds", duration)
        except Exception as e:
            raise InterfaceError(
                REST_ERROR, 'Error accessing {}.  ERROR:  {}'.format(url, e))
        if response.status < 300:
            return pulljson.JSONPullParser(
                HttpResponseAsUnicodeStream(response))
        if response.status < 400:
            raise InterfaceError(
                response.status,
                "HTTP Status: {}.   ERROR:  Redirection not supported.")
        else:
            msg = response.read().decode("utf8")
            try:
                errorDetails = json.loads(msg)
            except Exception:
                raise InterfaceError(
                    response.status, "HTTP Status: " + str(response.status) +
                    ", URL: " + url + ", Details:  " + str(msg))
            if response.status == HTTP_STATUS_DATABASE_ERROR:
                raise DatabaseError(
                    int(errorDetails['error']), errorDetails['message'])
            else:
                raise InterfaceError(response.status, "HTTP Status: " + str(
                    response.status) + ", URL: " + url +
                    ", Details:  " + str(errorDetails))


class HttpResponseAsUnicodeStream:

    def __init__(self, buf):
        self.stream = io.TextIOWrapper(
            HttpResponseIOWrapper(buf), encoding="utf8")

    def read(self, size):
        data = ""
        if not self.stream.closed:
            data = self.stream.read(size)
        return data


class HttpResponseIOWrapper:

    def __init__(self, buf):
        self.buf = buf
        self.closed = False

    def readable(self):
        return True

    def writable(self):
        return False

    def seekable(self):
        return False

    def read1(self, n=-1):
        return self.read(n)

    def read(self, size):
        return self.buf.read(size)



"""A pull parser for parsing JSON streams"""

# The MIT License (MIT)
#
# Copyright (c) 2015 by Teradata
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import sys
import decimal
import re
import json
import logging
from . import util  # @UnusedImport # noqa
if sys.version_info[0] == 2:
    from StringIO import StringIO  # @UnresolvedImport #@UnusedImport
else:
    from io import StringIO  # @UnresolvedImport @UnusedImport @Reimport # noqa

logger = logging.getLogger(__name__)

# JSONNode and value types.
OBJECT = "OBJECT"
ARRAY = "ARRAY"
FIELD = "FIELD"
STRING = "STRING"
NUMBER = "NUMBER"
BOOLEAN = "BOOLEAN"
NULL = "null"
TRUE = "true"
FALSE = "false"

# JSONEvent types.
START_OBJECT = "START_OBJECT"
START_ARRAY = "START_ARRAY"
FIELD_NAME = "FIELD_NAME"
FIELD_VALUE = "FIELD_VALUE"
ARRAY_VALUE = "ARRAY_VALUE"
END_OBJECT = "END_OBJECT"
END_ARRAY = "END_ARRAY"

# JSONParseError codes
JSON_SYNTAX_ERROR = "JSON_SYNTAX_ERROR"
JSON_INCOMPLETE_ERROR = "JSON_INCOMPLETE_ERROR"
JSON_UNEXPECTED_ELEMENT_ERROR = "JSON_UNEXPECTED_ELEMENT_ERROR"


class JSONPullParser (object):

    def __init__(self, stream, size=2 ** 16):
        """Initialize pull parser with a JSON stream."""
        self.stream = stream
        self.size = size
        self.node = None
        self.value = ""
        self.valueType = None
        self.tokens = []
        self.tokenIndex = 0
        self.halfToken = ""
        self.pattern = re.compile('([\[\]{}:\\\\",])')

    def expectObject(self):
        """Raise JSONParseError if next event is not the start of an object."""
        event = self.nextEvent()
        if event.type != START_OBJECT:
            raise JSONParseError(
                JSON_UNEXPECTED_ELEMENT_ERROR,
                "Expected START_OBJECT but got: " + str(event))

    def expectArray(self):
        """Raise JSONParseError if next event is not the start of an array."""
        event = self.nextEvent()
        if event.type != START_ARRAY:
            raise JSONParseError(
                JSON_UNEXPECTED_ELEMENT_ERROR,
                "Expected START_ARRAY but got: " + str(event))
        return JSONArrayIterator(self)

    def expectField(self, expectedName, expectedType=None, allowNull=False,
                    readAll=False):
        """Raise JSONParseError if next event is not the expected field with
           expected type else return the field value. If the next field is
           an OBJECT or ARRAY, only return whole object or array if
           readAll=True."""
        event = self.nextEvent()
        if event.type != FIELD_NAME:
            raise JSONParseError(
                JSON_UNEXPECTED_ELEMENT_ERROR,
                "Expected FIELD_NAME but got: " + str(event))
        if event.value != expectedName:
            raise JSONParseError(JSON_UNEXPECTED_ELEMENT_ERROR, "Expected " +
                                 expectedName + " field but got " +
                                 event.value + " instead.")
        return self._expectValue(FIELD_VALUE, expectedType, allowNull, readAll)

    def expectArrayValue(self, expectedType=None, allowNull=False,
                         readAll=False):
        """Raise JSONParseError if next event is not an array element with
           the expected type else return the field value. If the next value
           is an OBJECT or ARRAY, only return whole object or array if
           readAll=True."""
        return self._expectValue(ARRAY_VALUE, expectedType, allowNull, readAll)

    def _expectValue(self, eventType, expectedType, allowNull, readAll):
        event = self.nextEvent()
        if event.type == eventType:
            if allowNull and event.valueType == NULL:
                return None
            elif expectedType is not None and event.valueType != expectedType:
                raise JSONParseError(
                    JSON_UNEXPECTED_ELEMENT_ERROR, "Expected " + expectedType +
                    " but got " + event.valueType + " instead.")
            else:
                return event.value
        else:
            if eventType == ARRAY_VALUE:
                if event.node.parent is None or event.node.parent != ARRAY:
                    raise JSONParseError(
                        JSON_UNEXPECTED_ELEMENT_ERROR,
                        "Expected array element but not in an array.")
            if event.type == START_OBJECT:
                if expectedType is not None and expectedType != OBJECT:
                    raise JSONParseError(
                        JSON_UNEXPECTED_ELEMENT_ERROR, "Expected " +
                        expectedType + " but got an object instead.")
                elif expectedType is None or readAll:
                    return self.readObject(event)
            elif event.type == START_ARRAY:
                if expectedType is not None and expectedType != ARRAY:
                    raise JSONParseError(
                        JSON_UNEXPECTED_ELEMENT_ERROR, "Expected " +
                        expectedType + " but got array instead.")
                if expectedType is None or readAll:
                    return self.readArray(event)
                else:
                    return JSONArrayIterator(self)
            else:
                raise JSONParseError(
                    JSON_UNEXPECTED_ELEMENT_ERROR,
                    "Unexpected event: " + str(event))

    def readObject(self, event=None):
        """Read and return a JSON object."""
        if event is None:
            event = self.nextEvent()
            popRequired = False
        else:
            popRequired = True
        if event is None:
            return None
        if event.type != START_OBJECT:
            raise JSONParseError(
                JSON_UNEXPECTED_ELEMENT_ERROR,
                "Expected START_OBJECT but got " + event.type + " instead.")
        obj = self._load(event)
        if popRequired:
            self._pop()
        return obj

    def readArray(self, event=None):
        """Read and return a JSON array."""
        if event is None:
            event = self.nextEvent()
            popRequired = False
        else:
            popRequired = True
        if event is None:
            return None
        if event.type != START_ARRAY:
            raise JSONParseError(
                JSON_UNEXPECTED_ELEMENT_ERROR,
                "Expected START_ARRAY but got " + event.type + " instead.")
        arr = self._load(event)
        if popRequired:
            self._pop()
        return arr

    def nextEvent(self):
        """Iterator method, return next JSON event from the stream, raises
        StopIteration() when complete."""
        try:
            return self.__next__()
        except StopIteration:
            return None

    def next(self):
        """Iterator method, return next JSON event from the stream, raises
        StopIteration() when complete."""
        return self.__next__()

    def __next__(self):
        """Iterator method, return next JSON event from the stream, raises
        StopIteration() when complete."""
        while True:
            try:
                token = self.tokens[self.tokenIndex]
                self.tokenIndex += 1
                if token == "" or token.isspace():
                    pass
                elif token == '{':
                    return self._push(OBJECT)
                elif token == '}':
                    if self.node.type == FIELD:
                        self.tokenIndex -= 1
                        event = self._pop()
                        if event is not None:
                            return event
                    elif self.node.type == OBJECT:
                        return self._pop()
                    else:
                        raise JSONParseError(
                            JSON_SYNTAX_ERROR,
                            "A closing curly brace ('}') is only expected "
                            "at the end of an object.")
                elif token == '[':
                    if self.node is not None and self.node.type == OBJECT:
                        raise JSONParseError(
                            JSON_SYNTAX_ERROR, "An array in an object must "
                            "be preceded by a field name.")
                    return self._push(ARRAY)
                elif token == ']':
                    if self.valueType is not None:
                        self.tokenIndex -= 1
                        event = self._arrayValue()
                        if event is not None:
                            return event
                    elif self.node.type == ARRAY:
                        if self.node.lastIndex == self.node.arrayLength:
                            self.node.arrayLength += 1
                        return self._pop()
                    else:
                        raise JSONParseError(
                            JSON_SYNTAX_ERROR, "A closing bracket (']') "
                            "is only expected at the end of an array.")
                elif token == ':':
                    if self.node.type == OBJECT:
                        if self.value != "" and self.valueType == STRING:
                            event = self._push(FIELD, self.value)
                            self.value = ""
                            self.valueType = None
                            return event
                        else:
                            raise JSONParseError(
                                JSON_SYNTAX_ERROR,
                                "Name for name/value pairs cannot be empty.")
                    else:
                        raise JSONParseError(
                            JSON_SYNTAX_ERROR,
                            "A colon (':') can only following a field "
                            "name within an object.")
                elif token == ',':
                    if self.node.type == ARRAY:
                        event = self._arrayValue()
                        self.node.arrayLength += 1
                    elif self.node.type == FIELD:
                        event = self._pop()
                    else:
                        raise JSONParseError(
                            JSON_SYNTAX_ERROR,
                            "A comma (',') is only expected between fields "
                            "in objects or elements of an array.")
                    if event is not None:
                        return event
                else:
                    if self.valueType is not None:
                        raise JSONParseError(
                            JSON_SYNTAX_ERROR, "Extra name or value found "
                            "following: " + str(self.value))
                    elif self.node is None:
                        raise JSONParseError(
                            JSON_SYNTAX_ERROR,
                            "Input must start with either an "
                            "OBJECT ('{') or ARRAY ('['), got '" + token +
                            "' instead.")
                    elif token == '"':
                        escape = False
                        while True:
                            try:
                                token = self.tokens[self.tokenIndex]
                                self.tokenIndex += 1
                                if token == "":
                                    pass
                                elif escape:
                                    escape = False
                                    self.value += token
                                elif token == '"':
                                    break
                                elif token == '\\':
                                    escape = True
                                else:
                                    self.value += token
                            except IndexError:
                                data = self.stream.read(self.size)
                                if data == "":
                                    raise JSONParseError(
                                        JSON_INCOMPLETE_ERROR,
                                        "Reached end of input before " +
                                        "reaching end of string.")
                                self.tokens = self.pattern.split(data)
                                self.tokenIndex = 0
                        self.valueType = STRING
                    else:
                        token = token.strip()
                        if self.tokenIndex == len(self.tokens):
                            self.halfToken = token
                            raise IndexError
                        elif token[0].isdigit() or token[0] == '-':
                            self.value = decimal.Decimal(token)
                            self.valueType = NUMBER
                        elif token == "null":
                            self.value = None
                            self.valueType = NULL
                        elif token == "true":
                            self.value = True
                            self.valueType = BOOLEAN
                        elif token == "false":
                            self.value = False
                            self.valueType = BOOLEAN
                        else:
                            raise JSONParseError(
                                JSON_SYNTAX_ERROR,
                                "Unexpected token: " + token)
            except IndexError:
                data = self.stream.read(self.size)
                if data == "":
                    if self.node is not None:
                        raise JSONParseError(
                            JSON_INCOMPLETE_ERROR, "Reached end of input "
                            "before reaching end of JSON structures.")
                    else:
                        raise StopIteration()
                    return None
                logger.trace(data)
                self.tokens = self.pattern.split(data)
                self.tokenIndex = 0
                if self.halfToken is not None:
                    self.tokens[0] = self.halfToken + self.tokens[0]
                    self.halfToken = None

    def _load(self, event):
        if event.type == START_OBJECT:
            value = start = "{"
            end = "}"
        elif event.type == START_ARRAY:
            value = start = "["
            end = "]"
        else:
            raise JSONParseError(
                JSON_UNEXPECTED_ELEMENT_ERROR,
                "Unexpected event: " + event.type)
        count = 1
        tokens = self.tokens
        tokenIndex = self.tokenIndex
        inString = False
        inEscape = False
        try:
            while True:
                startIndex = tokenIndex
                for token in tokens[startIndex:]:
                    tokenIndex += 1
                    if token == "":
                        pass
                    elif inString:
                        if inEscape:
                            inEscape = False
                        elif token == '"':
                            inString = False
                        elif token == '\\':
                            inEscape = True
                    elif token == '"':
                        inString = True
                    elif token == start:
                        count += 1
                    elif token == end:
                        count -= 1
                        if count == 0:
                            value += "".join(tokens[startIndex:tokenIndex])
                            raise StopIteration()
                value += "".join(tokens[startIndex:])
                data = self.stream.read(self.size)
                if data == "":
                    raise JSONParseError(
                        JSON_INCOMPLETE_ERROR, "Reached end of input before "
                        "reaching end of JSON structures.")
                tokens = self.pattern.split(data)
                tokenIndex = 0
        except StopIteration:
            pass
        self.tokens = tokens
        self.tokenIndex = tokenIndex
        try:
            return json.loads(value, parse_float=decimal.Decimal,
                              parse_int=decimal.Decimal)
        except ValueError as e:
            raise JSONParseError(JSON_SYNTAX_ERROR, "".join(e.args))

    def _push(self, nodeType, value=None):
        if self.node is not None and self.node.type == FIELD:
            self.node.valueType = nodeType
        self.node = JSONNode(self.node, nodeType, value)
        if self.node.parent is not None and self.node.parent.type == ARRAY:
            self.node.arrayIndex = self.node.parent.arrayLength
            if self.node.parent.lastIndex == self.node.parent.arrayLength:
                raise JSONParseError(
                    JSON_SYNTAX_ERROR,
                    "Missing comma separating array elements.")
            self.node.parent.lastIndex = self.node.parent.arrayLength
        return self.node.startEvent()

    def _pop(self):
        # Pop the current node from the stack.
        node = self.node
        self.node = self.node.parent
        # Set the value and value type on the node.
        if node.valueType is None:
            node.valueType = self.valueType
            node.value = self.value
        # Reset value and valueType
        self.value = ""
        self.valueType = None
        if node.type == FIELD and node.valueType is None:
            raise JSONParseError(
                JSON_SYNTAX_ERROR, "Expected value for field: " + node.name)
        # Return the end event for the node.
        return node.endEvent()

    def _arrayValue(self):
        endOfArray = self.node.lastIndex == self.node.arrayLength
        if self.valueType is None and endOfArray:
            pass
        elif self.valueType is None:
            raise JSONParseError(
                JSON_SYNTAX_ERROR,
                "Expected value for array element at index: " +
                str(self.node.arrayLength))
        else:
            event = JSONEvent(
                self.node, ARRAY_VALUE, self.value, self.valueType,
                self.node.arrayLength)
            self.node.lastIndex = self.node.arrayLength
            # Reset value and valueType
            self.value = ""
            self.valueType = None
            # Return the end event for the node.
            return event

    def __iter__(self):
        return self

# Define exceptions


class JSONParseError(Exception):

    def __init__(self, code, msg):
        self.args = (code, msg)
        self.code = code
        self.msg = msg


class JSONNode (object):

    def __init__(self, parent, nodeType, name=None, value=None,
                 valueType=None):
        self.parent = parent
        self.type = nodeType
        self.name = name
        self.value = value
        self.valueType = valueType
        self.arrayIndex = None
        self.arrayLength = None
        self.lastIndex = -1
        if nodeType == ARRAY:
            self.arrayLength = 0

    def startEvent(self):
        if self.type == ARRAY:
            return JSONEvent(self, START_ARRAY, arrayIndex=self.arrayIndex)
        elif self.type == OBJECT:
            return JSONEvent(self, START_OBJECT, arrayIndex=self.arrayIndex)
        elif self.type == FIELD:
            return JSONEvent(self, FIELD_NAME, self.name)

    def endEvent(self):
        if self.type == ARRAY:
            return JSONEvent(self, END_ARRAY, arrayIndex=self.arrayIndex,
                             arrayLength=self.arrayLength)
        elif self.type == OBJECT:
            return JSONEvent(self, END_OBJECT, arrayIndex=self.arrayIndex)
        elif self.type == FIELD and self.valueType not in (OBJECT, ARRAY):
            return JSONEvent(self, FIELD_VALUE, self.value, self.valueType)


class JSONEvent (object):

    def __init__(self, node, eventType, value=None, valueType=None,
                 arrayIndex=None, arrayLength=None):
        self.node = node
        self.type = eventType
        self.value = value
        self.valueType = valueType
        self.arrayIndex = arrayIndex
        self.arrayLength = arrayLength

    def __repr__(self):
        text = "JSONEvent (type=" + self.type
        if self.value is not None:
            text += ", value=" + str(self.value)
        if self.valueType is not None:
            text += ", valueType=" + str(self.valueType)
        if self.arrayIndex is not None:
            text += ", arrayIndex=" + str(self.arrayIndex)
        if self.arrayLength is not None:
            text += ", arrayLength=" + str(self.arrayLength)
        text += ")"
        return text


class JSONArrayIterator (object):

    def __init__(self, parser):
        self.parser = parser
        self.complete = False

    def __iter__(self):
        return self

    def __next__(self):
        if self.complete:
            raise StopIteration()
        else:
            event = self.parser.nextEvent()
            if event.type == START_OBJECT:
                return self.parser.readObject(event)
            elif event.type == START_ARRAY:
                return self.parser.readArray(event)
            elif event.type == ARRAY_VALUE:
                return event.value
            elif event.type == END_ARRAY:
                self.complete = True
                raise StopIteration()
            else:
                raise JSONParseError(
                    JSON_UNEXPECTED_ELEMENT_ERROR,
                    "Unexpected event: " + str(event))

    def next(self):
        return self.__next__()
